// Code generated by mkpreempt.go; DO NOT EDIT.

package runtime

type xRegs struct {
	Z0  [64]byte
	Z1  [64]byte
	Z2  [64]byte
	Z3  [64]byte
	Z4  [64]byte
	Z5  [64]byte
	Z6  [64]byte
	Z7  [64]byte
	Z8  [64]byte
	Z9  [64]byte
	Z10 [64]byte
	Z11 [64]byte
	Z12 [64]byte
	Z13 [64]byte
	Z14 [64]byte
	Z15 [64]byte
	Z16 [64]byte
	Z17 [64]byte
	Z18 [64]byte
	Z19 [64]byte
	Z20 [64]byte
	Z21 [64]byte
	Z22 [64]byte
	Z23 [64]byte
	Z24 [64]byte
	Z25 [64]byte
	Z26 [64]byte
	Z27 [64]byte
	Z28 [64]byte
	Z29 [64]byte
	Z30 [64]byte
	Z31 [64]byte
	K0  uint64
	K1  uint64
	K2  uint64
	K3  uint64
	K4  uint64
	K5  uint64
	K6  uint64
	K7  uint64
}
