// Copyright 2025 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Issue 74648: wrong jump target when using PCALIGN.

package main

func F()

func main() {
	F()
}
