// Code generated by mkpreempt.go; DO NOT EDIT.

package runtime

type xRegs struct {
	V0  [16]byte
	V1  [16]byte
	V2  [16]byte
	V3  [16]byte
	V4  [16]byte
	V5  [16]byte
	V6  [16]byte
	V7  [16]byte
	V8  [16]byte
	V9  [16]byte
	V10 [16]byte
	V11 [16]byte
	V12 [16]byte
	V13 [16]byte
	V14 [16]byte
	V15 [16]byte
	V16 [16]byte
	V17 [16]byte
	V18 [16]byte
	V19 [16]byte
	V20 [16]byte
	V21 [16]byte
	V22 [16]byte
	V23 [16]byte
	V24 [16]byte
	V25 [16]byte
	V26 [16]byte
	V27 [16]byte
	V28 [16]byte
	V29 [16]byte
	V30 [16]byte
	V31 [16]byte
}
